/*
 * Copyright 1994, 1995 Cartoactive Systems, Vectaport Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the names of the copyright holders not be used in
 * advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  The copyright holders make
 * no representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 * FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 * 
 */

#include <MapUnidraw/mapcomps.h>
#include <MapUnidraw/mapeditor.h>
#include <MapUnidraw/mapviewer.h>
#include <MapUnidraw/mapviewstate.h>

#include <OverlayUnidraw/ovdamage.h>
#include <OverlayUnidraw/ovpage.h>
#include <OverlayUnidraw/ovselection.h>
#include <OverlayUnidraw/ovviews.h>
#include <Unidraw/Graphic/graphic.h>

#include <Attribute/attribute.h>
#include <Attribute/attrlist.h>
#include <Attribute/attrvalue.h>

#include <Unidraw/iterator.h>

#include <InterViews/cursor.h>
#include <InterViews/transformer.h>
#include <InterViews/window.h>
#include <IV-2_6/InterViews/painter.h>

#include <stdio.h>

/*****************************************************************************/

MapViewer::MapViewer (
    Editor* ed, GraphicView* gv, UPage* page, Grid* grid, 
    Coord w, Coord h, Orientation orientation,
    Alignment align, Zooming zoom
) : FrameViewer(ed, gv, page, grid, w, h, orientation, align, zoom)
{
}

float MapViewer::LimitMagnification (float desired) {
    const float lo = 1./512.;
    const float hi = 512.;

    return (desired < lo) ? lo : (desired > hi) ? hi : desired;
}

void MapViewer::Draw() 
{
    OverlaySelection* s = (OverlaySelection*)GetSelection();

    _editor->GetWindow()->cursor(hourglass);
    StartBuffering();
    s->ShowHighlights(this);
    GraphicBlock::Draw();
    FinishBuffering(true);

    s->Init(this);
    s->ShowHandles(this);
    _editor->GetWindow()->cursor(arrow);

    _damage->Reset();
}

void MapViewer::Redraw(Coord x0, Coord y0, Coord x1, Coord y1) 
{
    OverlaySelection* s = (OverlaySelection*)GetSelection();

    _editor->GetWindow()->cursor(hourglass);
    StartBuffering();
    s->ShowHighlights(this);
    GraphicBlock::Redraw(x0, y0, x1, y1);
    FinishBuffering(true);

    xorPainter->Clip(canvas, x0, y0, x1, y1);
    s->ShowHandles(this);
    xorPainter->NoClip();
    _editor->GetWindow()->cursor(arrow);
}

void MapViewer::GeoToDrawing(float lon, float lat, float& dx, float& dy) {
  MapEditor* ed = (MapEditor*)GetEditor();
  PJ* proj = ed->GetProjection();
  if (proj) {
    UV data;
    data.u = lon * DEG_TO_RAD;
    data.v = lat * DEG_TO_RAD;
    data = pj_fwd(data, proj);
    float px = data.u;
    float py = data.v;
    MapViewState* mvs = ed->GetMapViewState();
    if (mvs) {
      Transformer* t = mvs->deftrans();
      //float px, py;
      t->Transform(px, py, dx, dy);
    }
  }
  return;
}

void MapViewer::DrawingToGeo(float dx, float dy, float& lon, float& lat) {
  MapEditor* ed = (MapEditor*)GetEditor();
  MapViewState* mvs = ed->GetMapViewState();
  if (mvs) {
    Transformer* t = mvs->deftrans();
    float px, py;
    t->InvTransform(dx, dy, px, py);
    PJ* proj = ed->GetProjection();
    if (proj) { 
      UV data;
      data.u = px;
      data.v = py;
      data = pj_inv(data, proj);
      lon = data.u * RAD_TO_DEG;
      lat = data.v * RAD_TO_DEG;
    }
  }
  return;
}
void MapViewer::GeoToScreen(float lon, float lat, int& sx, int& sy) {
  float dx, dy;
  GeoToDrawing(lon, lat, dx, dy);
  DrawingToScreen(dx, dy, sx, sy);
}

void MapViewer::ScreenToGeo(int sx, int sy, float& lon, float& lat) {
  float dx, dy;
  ScreenToDrawing(sx, sy, dx, dy);
  DrawingToGeo(dx, dy, lon, lat);
}

void MapViewer::DrawingToUtm(float dx, float dy, float& utm_e, float& utm_n,
			     int& zone) {
  MapEditor* ed = (MapEditor*)GetEditor();
  MapViewState* mvs = ed->GetMapViewState();
  if (mvs) {
    Transformer* t = mvs->deftrans();
    t->inverse_transform(dx, dy, utm_e, utm_n);
  }
}

void MapViewer::UtmToDrawing(float utm_e, float utm_n, int zone,
			     float& dx, float& dy) {
  MapEditor* ed = (MapEditor*)GetEditor();
  MapViewState* mvs = ed->GetMapViewState();
  if (mvs) {
    Transformer* t = mvs->deftrans();
    t->transform(utm_e, utm_n, dx, dy);
  }
}
void MapViewer::ScreenToUtm(int sx, int sy, float& utm_e, float& utm_n) {
  float dx, dy;
  int zone;
  ScreenToDrawing(sx, sy, dx, dy);
  DrawingToUtm(dx, dy, utm_e, utm_n, zone);
}

void MapViewer::UtmToScreen(float utm_e, float utm_n, int zone,
			    int& sx, int& sy) {
  float dx, dy;
  UtmToDrawing(utm_e, utm_n, zone, dx, dy);
  DrawingToScreen(dx, dy, sx, sy);
}

void MapViewer::SetMapPage(OverlayComp* comp) {
  OverlayPage* page = make_map_page(comp);
  if (page) SetPage(page);
}

OverlayPage* MapViewer::make_map_page(OverlayComp* comp) {
  if (AttributeValue* sw_ev = comp->FindValue("sw_e")) 
    if (AttributeValue* sw_nv = comp->FindValue("sw_n"))
      if (AttributeValue* ne_ev = comp->FindValue("ne_e"))
	if (AttributeValue* ne_nv = comp->FindValue("ne_n")) {
	  
	  int ix1 = sw_ev->int_val();
	  int iy1 = sw_nv->int_val();
	  int ix2 = ne_ev->int_val();
	  int iy2 = ne_nv->int_val();
	  return new OverlayPage(ix2-ix1+1, iy2-iy1+1);
	}
  return nil;
}
